#version 120

/* DRAWBUFFERS:03 */

/*
const bool colortex0MipmapEnabled = true;
const bool colortex1MipmapEnabled = true;
const bool colortex2MipmapEnabled = true;
const bool colortex3MipmapEnabled = true;
*/

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D depthtex1;

uniform float frameTime;
uniform float viewWidth;
uniform float viewHeight;
uniform float smoothMask;
uniform float aspectRatio;
uniform float frameTimeCounter;

uniform int frameCounter;
uniform int isEyeInWater;

#define LENS_SETTINGS
#define BLOOM_TILE_LOAD

#include "/lib/shaderUtilities.glsl"

struct inputStruct {

    vec3 colorTexture0;

	float depthTexture1;

} Input;

#include "/lib/utility/misc/ditherIndexMap.glsl"
#include "/lib/utility/misc/computeRandomCoord.glsl"

#include "/lib/utility/post/fetchBloomTile.glsl"
#include "/lib/utility/post/computeDepthOfField.glsl"

void main() {

    Input.colorTexture0 = textureLoad(colortex0, texCoord).rgb;

	Input.depthTexture1 = textureLoad(depthtex1, texCoord).r;

	bool hand = textureLoad(colortex2, texCoord).g > 0;

	vec2 tileDistance = screenPixelsSize * 256.0 + 2.0;

	mat4x2 coords = mat4x2(0.0);

	coords[0] = upScale(texCoord, vec2(2.0), vec2(0.5));
	coords[1] = upScale(texCoord, vec2(2.0), vec2(1.0));

	coords[2] = upScale(texCoord, vec2(tileDistance.x, 2.0), vec2(0.0, 1.0));
	
	coords[3] = upScale(texCoord, tileDistance, vec2(tileDistance.x - 1.0, 0.0));

    vec3 tiledEffects = vec3(0.0);
	vec3 lensDirt = vec3(0.0);

	float maxLOD = computeMaxLOD(log2(screenPixels));

	float mixTime = (1.0 - frameTime);

	float smoothBuffer = luma(textureLoadLOD(colortex0, computeRandomCoord(texCoord), maxLOD).rgb);
	
	bool depthPixel = all(lessThanEqual(abs(texCoord - screenPixelsSize), screenPixelsSize));

	if (depthPixel) {

		smoothBuffer = textureLoad(depthtex1, vec2(0.5)).r;

		mixTime = sqrt(exp2(-mixTime) * phi - frameTime);

	}

	#ifdef LOAD_SMOOTH_BUFFERS
		smoothBuffer = mix(smoothBuffer, textureLoad(colortex3, texCoord).a, mixTime);
	#endif

	#ifdef ADVANCED_FILTERING
		float variableLOD = (1.0 - dither01) * 2;

		vec2 ditherOffset = dither01 + ivec2(1, -1);

		vec2 offsetSize = screenPixelsSize * 2;
	#else
	 	c(int) staticLOD = 2;
	#endif

	#ifdef BLOOM
		#ifdef LENS_DIRT
			#ifdef ADVANCED_FILTERING
				lensDirt = textureLoadLOD(colortex1, vec2( ditherOffset.x,  ditherOffset.x) * offsetSize + coords[2], variableLOD).rgb * 0.25 + lensDirt;
				lensDirt = textureLoadLOD(colortex1, vec2(-ditherOffset.y, -ditherOffset.y) * offsetSize + coords[2], variableLOD).rgb * 0.25 + lensDirt;
				lensDirt = textureLoadLOD(colortex1, vec2(-ditherOffset.y,  ditherOffset.x) * offsetSize + coords[2], variableLOD).rgb * 0.25 + lensDirt;
				lensDirt = textureLoadLOD(colortex1, vec2( ditherOffset.x, -ditherOffset.y) * offsetSize + coords[2], variableLOD).rgb * 0.25 + lensDirt;
			#else
				lensDirt = textureLoadLOD(colortex1, coords[2], staticLOD).rgb;
			#endif
		#else
			lensDirt = vec3(0.0);
		#endif
		
		lensDirt = gammaNonLinear(lensDirt);

		fetchBloomTile(tiledEffects, 0);
		fetchBloomTile(tiledEffects, 1);
		fetchBloomTile(tiledEffects, 2);
		
		if (BLOOM_TILES >= 4) fetchBloomTile(tiledEffects, 3);
		if (BLOOM_TILES >= 5) fetchBloomTile(tiledEffects, 4);
		if (BLOOM_TILES >= 6) fetchBloomTile(tiledEffects, 5);
		if (BLOOM_TILES >= 7) fetchBloomTile(tiledEffects, 6);

		tiledEffects = lensDirt * (tiledEffects * euler) + (tiledEffects * 0.0205078125);
	
		#if LENS_FLARE_TYPE > 0 && defined LENS_DIRT
			vec3 chromaHoop = vec3(0.0);

			mat3x2 chromaCoords = mat3x2(0.0);

			chromaCoords[0] = -texCoord + 1.0;

			chromaCoords[1] = fNormalize(0.5 - chromaCoords[0]);

			chromaCoords[2] = fract(chromaCoords[1] * 0.35 + chromaCoords[0]);

			vec2 tilePosition = bloomTilePositions[0] * (1.41421356 * screenPixelsSize + 1.0);

			vec2 disperse = screenPixelsSize * 16.0;

			chromaHoop.r = textureLoadLOD(colortex1, upScale(chromaCoords[1] * -disperse + chromaCoords[2], 4, tilePosition), 1).r;
			chromaHoop.g = textureLoadLOD(colortex1, upScale(chromaCoords[2] 							  , 4, tilePosition), 1).g;
			chromaHoop.b = textureLoadLOD(colortex1, upScale(chromaCoords[1] *  disperse + chromaCoords[2], 4, tilePosition), 1).b;

			chromaHoop = gammaNonLinear(chromaHoop * fLength(texCoord - 0.5) * sqrt(lensDirt));

			tiledEffects = chromaHoop * generate2DNoise(chromaCoords[1] * 512.0 + vec2(-time, time)) + tiledEffects;
		#endif
	#endif
	
	#ifdef LENS_DIRT
		#if LENS_FLARE_TYPE > 1 || defined LENS_DIFFRACTION
			vec3 advancedFlare = vec3(0.0);

			#ifdef ADVANCED_FILTERING
				advancedFlare = textureLoadLOD(colortex1, vec2( ditherOffset.x,  ditherOffset.x) * offsetSize + coords[3], variableLOD).rgb * 0.25 + advancedFlare;
				advancedFlare = textureLoadLOD(colortex1, vec2(-ditherOffset.y, -ditherOffset.y) * offsetSize + coords[3], variableLOD).rgb * 0.25 + advancedFlare;
				advancedFlare = textureLoadLOD(colortex1, vec2(-ditherOffset.y,  ditherOffset.x) * offsetSize + coords[3], variableLOD).rgb * 0.25 + advancedFlare;
				advancedFlare = textureLoadLOD(colortex1, vec2( ditherOffset.x, -ditherOffset.y) * offsetSize + coords[3], variableLOD).rgb * 0.25 + advancedFlare;
			#else
				advancedFlare = textureLoadLOD(colortex1, coords[3], staticLOD).rgb;
			#endif

			tiledEffects = gammaNonLinear(advancedFlare) + tiledEffects;
		#endif
	#endif

	#ifdef LENS_GLARE
		c(float) sampleCount = LENS_EFFECT_SAMPLES * 0.125;

		tiledEffects *= sampleCount * 2.0;

		sampleRCPVar(sampleCount);

		vec2 coordOffset = vec2(0.0);
		
		vec2 blurSize = screenPixelsSize * sampleRCP * 8.0;
		
		#ifdef ADVANCED_FILTERING
			blurSize *= dither01;

			variableLOD = variableLOD * 1.25 + 1.0;
		#endif

		for (int i = 1; i <= sampleCount; ++i) {

			coordOffset = vec2(0.0, i);

			if (LENS_GLARE_TYPE > 0)
			coordOffset = coordOffset.yx;
			
			#ifdef ADVANCED_FILTERING
				tiledEffects = gammaNonLinear(textureLoadLOD(colortex1,  coordOffset * blurSize + coords[1], variableLOD).rgb) + tiledEffects;
				tiledEffects = gammaNonLinear(textureLoadLOD(colortex1, -coordOffset * blurSize + coords[1], variableLOD).rgb) + tiledEffects;
			#else
				tiledEffects = gammaNonLinear(textureLoadLOD(colortex1,  coordOffset * blurSize + coords[1], staticLOD).rgb) + tiledEffects;
				tiledEffects = gammaNonLinear(textureLoadLOD(colortex1, -coordOffset * blurSize + coords[1], staticLOD).rgb) + tiledEffects;
			#endif

		}

		tiledEffects *= sampleRCP * 0.5;
	#endif

	#ifdef DEPTH_OF_FIELD
		if (!hand) computeDepthOfField(Input.colorTexture0, Input.depthTexture1);
	#endif

	#ifdef AUTO_EXPOSURE
		#ifndef CAMERA_ARTIFACT_CORRECTION
			vec2 EV = vec2(textureLoad(colortex3, texCoord).a, 0.0);

			if (depthPixel) EV.x = textureLoad(colortex3, screenPixelsSize).a;
		#else
			vec2 EV = vec2(textureLoadLOD(colortex3, coords[0], maxLOD).a, 0.0);
		#endif
		
		EV.x = RCP(EV.x * 4.0 + rcp16bit);

		EV.y = max(EV.x - inversesqrt(EV.x + 1.0), luma(sunLuminance * 0.08 + tiledEffects) / sunLuminance);
		
		EV.x /= min0(log2(EV.x)) + 1.0;
	#else
		vec2 EV = vec2(1.0);

		EV.x = RCP((powEvenInt(APERTURE, 2) * SHUTTER_SPEED) / (ISO * 12.5));

		EV.y = max(EV.x - inversesqrt(EV.x + 1.0), luma(sunLuminance * 0.08 + tiledEffects) / sunLuminance);
	#endif

	EV = sqrt(EV * 0.5) / 0.8333;

	EV.y /= sqrt(powEvenInt(EV.x * 2.0, 2) * 0.8333 + 1.0);

	Input.colorTexture0 = (gammaLinear(tiledEffects) * EV.y + Input.colorTexture0) * EV.x;
	
    pixelData0 = vec4(Input.colorTexture0, !hand);
	pixelData1 = vec4(textureLoad(colortex3, texCoord).rgb, smoothBuffer);
}