#ifdef PIXEL
    uniform sampler2D texture;
    uniform sampler2D normals;

    #if defined TERRAIN || defined HAND || defined HAND_WATER
        varying mat4x3 vertexAttributes;
    #endif

    #ifdef TERRAIN
        varying mat3x2 parallaxAttributes;
    #endif

    varying vec4 colorInput;

    varying vec2 lightmapInput;

    #ifdef ENTITIES
        uniform vec4 entityColor;
    #endif

    #if defined HAND || defined HAND_WATER
        uniform int heldItemId;
    #endif

    #ifdef WATER 
        uniform int isEyeInWater;
    #endif

    uniform mat4 gbufferModelViewInverse;

    #include "/lib/shaderUtilities.glsl"
    
    #ifdef TERRAIN
        #include "/lib/utility/misc/terrainParallax.glsl"
    
        flat(mat3) tbnMatrix;
    #endif

    flat(vec3) normalInput;

    flat(float) IDInput;

    void main() {

        vec4 textureInput = texture2D(texture, texCoord) * colorInput;
        
        vec3 normalOutput = normalInput;

        vec2 lightmapOutput = lightmapInput;

        float layer = 1.0;
        float hand = 0.0;

        float IDs = round(IDInput);

        #ifdef ENTITIES
            textureInput.rgb = mix(textureInput.rgb, entityColor.rgb, entityColor.a);
        #endif

        #ifdef GLINT
            lightmapOutput = vec2(0.0);
            normalOutput = vec3(0.0);
        #endif

        #if defined HAND || defined HAND_WATER
            if (heldItemId < 256) 
            IDs = heldItemId;
            
            hand = 1.0;
        #endif

        #ifdef WATER
            if (waterSource) textureInput = vec4(exp2(-waterColor.rgb), waterColor.a);

            if (underWater) 
            layer = 0.0;
        #endif

        #if defined WATER || defined HAND_WATER
            if (glassSource) {
                
                #if GLASS_LAYERS == 0
                    textureInput.a = 1.0;
                #elif GLASS_LAYERS == 1
                    textureInput.a = floor(textureInput.a + 0.5) - 0.5;
                #endif

                layer = 0.0;

            }

            if (IDs == 90.0) layer = 0.0, textureInput.a = floor(textureInput.a + 0.5) - 0.5;
        #endif

        #ifdef BEACONBEAM
            IDs = 138;
            layer = 0.0;
        #endif

        #ifdef TERRAIN
            vec2 parallaxCoord = texCoord;

            mat2x2 wrapTexture = mat2x2(0.0);

            wrapTexture[0] = dFdx(texCoord);
            wrapTexture[1] = dFdy(texCoord);

            #ifdef PARALLAX_MAPPING
                parallaxCoord = computeParallaxCoord(wrapTexture);

                textureInput = texture2DGradARB(texture, parallaxCoord, wrapTexture[0], wrapTexture[1]) * colorInput;
            #endif

            #ifdef BUMP_MAPPING
                normalOutput = fNormalize(texture2DGradARB(normals, parallaxCoord, wrapTexture[0], wrapTexture[1]).rgb * 2.0 - 1.0) * tbnMatrix;
            #endif
        #endif
        
        IDs = min1(IDs);
        
        if (IDs == 7) IDs = 1;
        if (glassSource) IDs = 7;

        if (IDs == 6) IDs = 1;
        if (iceSource) IDs = 6;

        if (IDs == 5) IDs = 1;
        if (waterSource) IDs = 5;

        if (IDs == 4) IDs = 1;
        if (warmSource) IDs = 4;

        if (IDs == 3) IDs = 1;
        if (emissiveSource && !warmSource) IDs = 3;

        if (IDs == 2) IDs = 1;
        if (plantSource) IDs = 2;
        
        if (IDs > 7) IDs = 1;

        normalOutput = mat3(gbufferModelViewInverse) * normalOutput;
        
        pixelData0 = textureInput;
        pixelData1 = vec4(encodeNormal(normalOutput), encode2x8(lightmapOutput), encode2x4(vec2(IDs * 0.125, hand)), layer);
    }
#endif

#ifdef VERTEX
attribute vec4 mc_Entity;
attribute vec4 at_tangent;
attribute vec4 mc_midTexCoord;

    varying vec4 colorInput;

    varying vec2 lightmapInput;

    #if defined TERRAIN || defined HAND || defined HAND_WATER
        varying mat4x3 vertexAttributes;

        varying mat3x2 parallaxAttributes;

        uniform vec3 cameraPosition;

        uniform mat4 gbufferModelViewInverse;
        uniform mat4 gbufferModelView;

        uniform float frameTimeCounter;
    #endif

    #include "/lib/shaderUtilities.glsl"

    #if defined TERRAIN || defined HAND || defined HAND_WATER
        #include "/lib/utility/misc/displaceVertex.glsl"
    #endif

    #ifdef TERRAIN
        flat(mat3) tbnMatrix;
    #endif
    
    #if ANTI_ALIASING == 2 || ANTI_ALIASING == 3
        uniform int frameCounter;

        uniform float viewWidth;
        uniform float viewHeight;
        uniform float frameTime;

        vec2 temporalJitter() {
            return circleMap((frameCounter & int(0.0625 / frameTime - 1)) + 1) / vec2(viewWidth, viewHeight); // Cycles per frame must be atleast 1/16th of the frame-rate for a stable image with very low flickering.     
        }
    #endif
    
    flat(vec3) normalInput;

    flat(float) IDInput;

    void main() {

        texCoord      = gl_MultiTexCoord0.xy;
        lightmapInput = gl_MultiTexCoord1.xy * rcp8bit + rcp8bit;
        normalInput   = gl_NormalMatrix * gl_Normal;
        colorInput    = gl_Color;
        IDInput       = mc_Entity.x;

        #ifdef TERRAIN       
            parallaxAttributes[1] = mc_midTexCoord.xy;

            vec2 midCoord = texCoord - mc_midTexCoord.xy;

            parallaxAttributes[2] = fSign(midCoord) * 0.5 + 0.5;
            parallaxAttributes[0] = abs(midCoord) * 2.0;
            parallaxAttributes[1] = min(texCoord, parallaxAttributes[1] - midCoord);
        #endif
        
        mat2x3 position = mat2x3(0.0);

        #if defined TERRAIN || defined HAND || defined HAND_WATER
            position[0] = transMAD(gbufferModelViewInverse, transMAD(gl_ModelViewMatrix, gl_Vertex.xyz));

            position[1] = position[0].xyz + cameraPosition;

            #ifdef VERTEX_DISPLACEMENT
                #ifdef TERRAIN
                    if (
                    IDInput == 141.0 ||
                    IDInput == 142.0 ||
                    IDInput == 115.0 ||
                    IDInput == 105.0 ||
                    IDInput == 104.0 ||
                    IDInput == 59.0  ||
                    IDInput == 51.0	 ||
                    IDInput == 40.0  ||
                    IDInput == 39.0  ||
                    IDInput == 38.0  ||
                    IDInput == 37.0	 ||
                    IDInput == 31.0  ||
                    IDInput == 32.0  ||
                    IDInput == 30.0	 ||
                    IDInput == 6.0	 )
                    displaceVertex(position[0], lightmapInput.y, 0.75, true);

                    if (
                    IDInput == 161.0 || 
                    IDInput == 106.0 ||
                    IDInput == 18.0  )
                    displaceVertex(position[0], lightmapInput.y, 0.25, false);

                    if (
                    IDInput == 175.0 ||
                    IDInput == 83.0  )
                    displaceVertex(position[0], lightmapInput.y, 0.1, false);
    
                    if (
                    IDInput == 141.0 ||
                    IDInput == 142.0 ||
                    IDInput == 115.0 ||
                    IDInput == 59.0  ||
                    IDInput == 38.0  ||
                    IDInput == 37.0	 ||
                    IDInput == 31.0  &&
                    fLength(position[0]) < 2.0 && texCoord.y < mc_midTexCoord.y)
                    position[0].xz *= 2.0 - clamp01(fLength(position[0] - vec3(0.0, 0.5, 0.0)));
                #endif

                #ifdef HAND
                    vec4 waves = vec4(0.0);
                    
                    c(vec3) power = vec3(0.00025, 0.00075, 0.0025);

                    waves.x = sin(time);
                    waves.y = cos(time);

                    waves.z = waves.x * power.z;

                    waves.xyz = waves.xyz * power;

                    waves.w = dot(waves.xyz, power) * dotX(RCP(power));
      
	                position[0] = waves.xyz * waves.w + position[0];
                #endif    
            #endif

            vertexAttributes[0] = position[1].xyz;

            position[0] = transMAD(gbufferModelView, position[0]);

            gl_Position = diagonal4(gl_ProjectionMatrix);
            
            gl_Position.xyw = position[0].xyz * gl_Position.xyw;

            gl_Position.z = position[0].z * gl_Position.z + gl_ProjectionMatrix[3].z;

            #ifdef TERRAIN
                vertexAttributes[3] = fNormalize(position[0]);
                vertexAttributes[2] = fNormalize(gl_NormalMatrix * at_tangent.xyz);
                vertexAttributes[1] = cross(vertexAttributes[2], normalInput);

                tbnMatrix = transpose(mat3(vertexAttributes[2], vertexAttributes[1], normalInput));
                    
                vertexAttributes[3] = tbnMatrix * vertexAttributes[3];
            #endif
        #else
            position[0] = transMAD(gl_ModelViewMatrix, gl_Vertex.xyz);

            gl_Position = diagonal4(gl_ProjectionMatrix);
            
            gl_Position.xyw = position[0].xyz * gl_Position.xyw;

            gl_Position.z = position[0].z * gl_Position.z + gl_ProjectionMatrix[3].z;
        #endif

        #if ANTI_ALIASING == 2 || ANTI_ALIASING == 3
            gl_Position.xy = temporalJitter() * gl_Position.w + gl_Position.xy;
        #endif
        
    }
#endif