varying vec2 texCoord;

#include "/lib/asset/settings.glsl"

#include "/lib/asset/values.glsl"

#include "/lib/asset/macros.glsl"

#include "/lib/asset/helpers.glsl"

#if !defined VERTEX && !defined PIXEL
    vec2 screenPixels = vec2(viewWidth, viewHeight);

    vec2 screenPixelsSize = RCP(screenPixels);
#endif

#ifdef UP_DIRECTION
    vec3 upDirection = fNormalize(upPosition);
#endif

#ifdef LIGHT_MATRIX
    mat2x3 lightPosition = mat2x3(sunPosition * 0.01, shadowLightPosition * 0.01);
#endif

#ifdef SHADOW_MATRIX
    mat4 shadowProjMatrix = shadowProjection * shadowModelView;
#endif

#ifdef FOV_PROJECTION
    float FOV = underWater ? gbufferProjection[1][1] * tan(atan(gbufferProjectionInverse[1][1]) * 0.85) : 1.0;
#endif

#if defined BLOOM_TILE_LOAD || defined BLOOM_TILE_COMPUTE
    c(ivec2) bloomTilePositions[7] = ivec2[7](
    ivec2(0, 0),
    ivec2(2, 0),
    ivec2(6, 0),
    ivec2(0, 8),
    ivec2(2, 16),
    ivec2(6, 32),
    ivec2(14, 64));
#endif

#ifdef LENS_SETTINGS
    uniform mat4 gbufferProjectionInverse;

    c(vec3) internalScale = vec3(16.0 - BLADES, fPow(cos(pi / BLADES), 1.0 - ROUNDING), pi * BLADES);

    c(int) sampleLOD = 32 / LENS_EFFECT_SAMPLES;
    
    float viewDiag = sqrt(dot(fPow(screenPixels, 2.0), vec2(1.0)));

	float focalLengthMM = (viewDiag / (2.0 * (pi * atan(gbufferProjectionInverse[1][1]) / 360.0))) / (viewDiag * 1000.0) / RINGS;
    
    float fStop = focalLengthMM / APERTURE;
    
    #ifdef REFRACTION_CHROMATIC_DISPERSION
        float disperseDifference = exp2(-sqrt(fStop)) * sqrt(fStop);
        vec3 dispersion = vec3(1 - disperseDifference, 1, 1 + disperseDifference);
    #else
        c(vec3) dispersion = vec3(1.0, 1.0, 1.0);
    #endif

    vec2 bokehPointMap(in float point, in float progress, c(in float) piScaleBokeh, c(in float) tauScale) {
        return min(progress * fPow(abs(cos(point * piScaleBokeh)), internalScale.x) + internalScale.y, progress) * circleMap(point * tauScale + ROTATION);
    }
#endif