float computeFocus(in float focusPoint, in float depthInput) {
	
	depthInput = 0.1 / (1.1 - depthInput);

	return ((fStop * (focusPoint - depthInput)) / (focusPoint * (depthInput - fStop)));

}

void computeDepthOfField(io vec3 output0, in float depthInput) {

	float colorCorrect = 0.0;

	float distanceLimit = focalLengthMM * 0.5;

	#ifdef AUTO_FOCUS
		float focusPoint = 0.1 / (1.1 - textureLoad(colortex3, screenPixelsSize).a);
	#else
		c(float) focusPoint = 0.1 / (1.1 - DEPTH_VALUE);
	#endif

	vec4 focus = vec4(0.0);
	
	focus.z = computeFocus(focusPoint, depthInput);

	focus.w = clamp(smoothMask * distanceLimit + focus.z, -distanceLimit, distanceLimit);

	#ifdef BARREL_DISTORTION
		vec2 centerDist = (texCoord - 0.5) * 2.0;

		float angle = 1.0 - cos(fLength(centerDist));
		
		vec2 rotationDist = vec2(centerDist.y, -centerDist.x) * angle;

		focus.w = angle * distanceLimit * 0.5 * fSign(focus.w) + focus.w;
	#endif

	#ifdef TILT_SHIFT
		focus.w = abs(texCoord.y - 0.5) * fSign(focus.w) * distanceLimit + focus.w;
	#endif

	focus.xy = focus.w * vec2(0.5, aspectRatio * 0.5);

	vec3 totalSamples = vec3(0.0);

	mat3x2 sampleCoord = mat3x2(texCoord, texCoord, texCoord);

	#ifdef REFRACTION_CHROMATIC_ABERRATION
		float chromaticFocus = abs(focus.w) * 0.5;

		sampleCoord[0] = rotateVector(sampleCoord[0], -chromaticFocus);
		sampleCoord[2] = rotateVector(sampleCoord[2],  chromaticFocus);
	#endif

	focus.w = sqrt(abs(focus.w));

	#ifdef DEPTH_OF_FIELD_SAMPLE_CUTOFF
		if (!(focus.w * 8.0 - sqrt(focus.w) > 0)) return;
	#endif

	#ifdef DEPTH_OF_FIELD_VARIABLE_SAMPLES
		float sampleCount = min(powEvenInt(LENS_EFFECT_SAMPLES, 2) * focus.w, LENS_EFFECT_SAMPLES);
		
		float sampleRCP = RCP(sampleCount);

		float tauScale = tau * sampleRCP;
		float bokehPiScale = (internalScale.z + ROTATION) * sampleRCP;
	#else
		c(float) sampleCount = LENS_EFFECT_SAMPLES;
		
		sampleRCPVar(sampleCount);

		c(float) tauScale = tau * sampleRCP;
		c(float) bokehPiScale = (internalScale.z + ROTATION) * sampleRCP;
	#endif
	
	float LOD = ((sampleLOD * focus.w) * 8.0 - sqrt(focus.w)) * 2.0;

	for (int i = 1; i <= RINGS; ++i) 

	for (int j = 0; j < i * sampleCount; ++j) {

		vec2 bokehPoints = bokehPointMap(j, i, bokehPiScale, tauScale) * focus.xy;

		#ifdef BARREL_DISTORTION
			bokehPoints = rotationDist * dot(rotationDist, bokehPoints) + bokehPoints;
		#endif

		#ifdef DEPTH_AWARE_BLUR
			focus.w = computeFocus(focusPoint, textureLoad(depthtex1, texCoord + bokehPoints).r);

			if (abs(focus.z - focus.w) > abs(focus.w)) continue;
		#endif

		if (textureLoadLOD(colortex2, bokehPoints + sampleCoord[1], LOD).g > 0) continue;

		totalSamples = vec3(
		textureLoadLOD(colortex0, bokehPoints * dispersion.x + sampleCoord[0], LOD).r,
		textureLoadLOD(colortex0, bokehPoints * dispersion.y + sampleCoord[1], LOD).g,
		textureLoadLOD(colortex0, bokehPoints * dispersion.z + sampleCoord[2], LOD).b) + totalSamples;

		++colorCorrect;

	}

	if (colorCorrect > 0)
	output0 = totalSamples / colorCorrect;

}