#define cloudCoverage2D(x, y) powEvenInt(clamp01(((y + x - 1.0)) * 1.1 - 0.1), 2)

float cloudFBM2D(in vec2 cloudPlane, out mat4x2 detailCoords, in vec3 weights, in float weight) {

	float cloudMovement = time * 0.02;

	detailCoords[0] = cloudPlane * 0.014;

	detailCoords[0] += generate2DNoise(detailCoords[0]) * 0.3 - 0.15;

	detailCoords[0].x = detailCoords[0].x * 0.25 + cloudMovement;
	
	float cloud = -generate2DNoise(detailCoords[0]);
	
	detailCoords[1] = detailCoords[0] * 2.0 - cloud * vec2(0.5, 1.35) + vec2(cloudMovement, 0.0);
	
	cloud = generate2DNoise(detailCoords[1]) * weights.x + cloud;
	
	detailCoords[2] = detailCoords[1] * vec2(12.0, 2.65) + cloudMovement * vec2(3.0, 0.55) - cloud * vec2(1.5, 0.75);
	
	cloud = generate2DNoise(detailCoords[2]) * weights.y + cloud;
	
	detailCoords[3] = detailCoords[2] * 3.0 + cloudMovement;
	
	cloud = generate2DNoise(detailCoords[3]) * weights.z + cloud;
	
	cloud = generate2DNoise(detailCoords[3] * 3.0 + cloudMovement) * 0.022 + cloud;

	return weight - cloud;
}

void computeTwoDimensionalClouds(io vec3 output0, in vec3 worldSpacePosition, in vec3 rayleigh, in vec2 horizon, in vec2 lightAura, io float cloudSurface) {

	c(float) cloudHeight = 384.0;

    bool aboveClouds = cameraPosition.y >= cloudHeight;

	if (aboveClouds) 
	horizon.y = 0.5;

	if (worldSpacePosition.y + cameraPosition.y < cloudHeight && horizon.x >= 0.1 || aboveClouds && worldSpacePosition.y < 0.0) {

		float yRCP = RCP(worldSpacePosition.y);

		worldSpacePosition.y = dotX(worldSpacePosition.xz) * (yRCP / 512) + worldSpacePosition.y;

		c(float) coverage = 0.67;
		c(float) cloudLight = 5.0;

		c(vec3) weights = vec3(0.472, 0.107, 0.0427);
		c(float) weight = weights.x + weights.y + weights.z;

		float visibility = (1.0 - coverage) * abs(worldSpacePosition.y + 3.0) * horizon.y;

		vec2 cloudPlane = worldSpacePosition.xz / worldSpacePosition.y * (cloudHeight - cameraPosition.y) + cameraPosition.xz;

		vec4 cloudOutput = vec4(0.0);
		mat4x2 detailCoords = mat4x2(0.0);

		cloudOutput.a = cloudCoverage2D(cloudFBM2D(cloudPlane, detailCoords, weights, weight) * coverage, coverage);

		vec2 lightOffset = vec2(dot(lightPosition[1], gbufferModelView[0].xyz), dot(lightPosition[1], gbufferModelView[2].xyz)) * 0.2;

		float detail = -generate2DNoise(detailCoords[0] + lightOffset);
		
		detail = generate2DNoise(detailCoords[1] + lightOffset) * weights.x + detail;
		detail = generate2DNoise(detailCoords[2] + lightOffset) * weights.y + detail;
		detail = generate2DNoise(detailCoords[3] + lightOffset) * weights.z + detail;

		detail = powEvenInt(1.5 - cloudCoverage2D(weight - detail, coverage), 6) * cloudLight;

		detail = clamp01(detail * powEvenInt(cloudOutput.a * 2, 2) + powEvenInt(lightAura.y * mix(0.1, 0.0, clamp01(timeNight + rainStrength)), 6));
		
		cloudOutput.rgb = mix(rayleigh, rayleigh * cloudLight, detail);
		
		output0 = mix(output0, gammaNonLinear(cloudOutput.rgb), cloudOutput.a * (1.0 - cloudSurface) * visibility * mix(1.0, lightAura.y * 0.05 + 0.01, timeNight));
		
		cloudSurface = clamp01(cloudOutput.a * cloudOutput.a + cloudSurface);

	}
}