float cloudMovement = time * 0.02;

c(vec3) weights = vec3(0.393, 0.069, 0.023);

c(float) weight = weights.x + weights.y + weights.z;

#define cloudCoverage3D(x, y) clamp01((y + x - 1.0) * 10.0 - 0.1)

float cloudFBM3D(in vec3 worldPosition) {

    float cloud = 0.0;
    
    worldPosition = worldPosition * 0.007;

    worldPosition.x = worldPosition.x * 0.6 + cloudMovement;

    worldPosition = worldPosition * vec3(1.0, 2.0, 2.0) + cloudMovement;

    cloud = generate3DNoise(cloud * vec3(2.0, 2.0, 2.0) + worldPosition) * weights.x + cloud;
    
    worldPosition = worldPosition * vec3(8.0, 6.0, 6.0) + cloudMovement;

    cloud = generate3DNoise(cloud * vec3(8.0, 6.0, 6.0) + worldPosition) * weights.y + cloud;

    worldPosition = worldPosition * vec3(5.0, 4.0, 4.0) + cloudMovement;

    cloud = generate3DNoise(cloud * vec3(5.0, 4.0, 4.0) + worldPosition) * weights.z + cloud;

    return weight - cloud;
    
}

void computeThreeDimensionalClouds(io vec3 output0, in vec3 worldSpacePosition, in vec3 rayleigh, in vec2 horizon, in vec2 lightAura, io float cloudSurface) {

    c(float) cloudLowerHeight = 256.0;
    c(float) cloudUpperHeight = 304.0;

    c(float) cloudDepth = cloudUpperHeight - cloudLowerHeight;
    
    mat2x3 cloudPlane = mat2x3(0.0);

    bool raymarch = true;
    bool aboveClouds = cameraPosition.y >= cloudLowerHeight;

    float yRCP = RCP(worldSpacePosition.y);

    worldSpacePosition.y = dotX(worldSpacePosition.xz) * (yRCP / 512) + worldSpacePosition.y;

    cloudPlane[0] = worldSpacePosition * ((cloudUpperHeight - cameraPosition.y) * yRCP);
    cloudPlane[1] = worldSpacePosition * ((cloudLowerHeight - cameraPosition.y) * yRCP);

    if (cameraPosition.y < cloudLowerHeight) {

        if (worldSpacePosition.y <= 0.0)
        
        raymarch = false;

        swap(cloudPlane[0], cloudPlane[1]);

    } else if (cloudLowerHeight <= cameraPosition.y && cameraPosition.y <= cloudUpperHeight) {

		if (worldSpacePosition.y < 0.0) 
        
        swap(cloudPlane[0],cloudPlane[1]);
        
		cloudPlane[1] = vec3(0.0);

		swap(cloudPlane[0], cloudPlane[1]);
            
	} 

    else 
    
    if (worldSpacePosition.y >= 0.0) 
    raymarch = false;
	
    if (aboveClouds) 
    horizon.y = 0.5;

    vec4 cloudOutput = vec4(0.0);

    if (raymarch && horizon.x >= 0.1 || raymarch && aboveClouds) {

        vec3 rayIncrement = (cloudPlane[1] - cloudPlane[0]) * 0.125;

        vec3 rayPosition = rayIncrement * (dither01 + 1.0) + (cloudPlane[0] + cameraPosition);

        c(float) coverage = 0.67;
        c(float) cloudCenter = cloudDepth * 0.5 + cloudLowerHeight;
        c(float) cloudDepthHalf = RCP(cloudDepth * 0.5);
        c(float) cloudFalloff = RCP(cloudDepth - cloudDepth * 0.25);
        c(float) cloudLight = 5.0;

        float lowlight = clamp01(timeNight + rainStrength);

        lightAura.x = lightAura.y * mix(2.0, 0.5, lowlight);
        lightAura.y = lightAura.y * mix(0.25, 0.0, lowlight);

        float weight = 0.0;
        float FBM = 0.0;

        for (float i = 0.0; i < 4 && cloudOutput.a < 1.0; ++i, rayPosition += rayIncrement) {
        
            weight = 1.5 - clamp01(abs(rayPosition.y - cloudCenter) * cloudDepthHalf);

            FBM = powEvenInt(cloudCoverage3D(coverage, cloudFBM3D(rayPosition) * weight) - weights.z, 2);
            
            FBM /= 0.5 + FBM * 0.5;

            weight = powEvenInt(abs(rayPosition.y - cloudLowerHeight) * cloudFalloff, 6) * powEvenInt(FBM * 2.5, 2) + lightAura.y;

            cloudOutput.rgb = mix(cloudOutput.rgb, mix(rayleigh, rayleigh * cloudLight, clamp01(weight)), (1.0 - cloudOutput.a) * FBM);

            cloudOutput.a = FBM + cloudOutput.a;
            
        }
        
        cloudOutput.a = clamp01(cloudOutput.a);
        
        output0 = mix(output0, gammaNonLinear(cloudOutput.rgb), cloudOutput.a * (1.0 - cloudSurface) * horizon.y * mix(1.0 - coverage, lightAura.x * 0.05 + 0.01, timeNight));

        cloudSurface = clamp01(cloudOutput.a * cloudOutput.a + cloudSurface);
    }
}
